# Copyright (c) 2021 Arista Networks, Inc.
# Use of this source code is governed by the Apache License 2.0
# that can be found in the COPYING file.

from cvprac.cvp_client import CvpClient
from pprint import pprint as pp
import ssl
ssl._create_default_https_context = ssl._create_unverified_context
import requests.packages.urllib3
requests.packages.urllib3.disable_warnings()

# Reading the service account token from a file
with open("token.tok") as f:
    token = f.read().strip('\n')

clnt = CvpClient()
clnt.connect(nodes=['cvp1'], username='',password='',api_token=token)

def get_events_all(client):
    ''' Get All events '''
    event_url = '/api/resources/event/v1/Event/all'
    response = client.get(event_url)
    return response['data']

def get_event(client, key, ts):
    event_url = '/api/resources/event/v1/Event?'
    url = event_url + 'key.key=' + key + "&key.timestamp=" + ts
    response = client.get(url)
    return response

def get_events_t1_t2(client, t1, t2):
    event_url = '/api/resources/event/v1/Event/all?'
    url = event_url + 'time.start=' + t1 + "&time.end=" + t2
    response = client.get(url)
    return response['data']

def get_events_by_severity(client, severity):
    payload = {"partialEqFilter": [{"severity": severity }]}
    event_url = '/api/resources/event/v1/Event/all'
    response = client.post(event_url, data=payload)
    if 'data' in response.keys():
        return response['data']
    else:
        return response

def get_events_by_type(client, etype):
    payload = {"partialEqFilter": [{"eventType": etype }]}
    event_url = '/api/resources/event/v1/Event/all'
    response = client.post(event_url, data=payload)
    if 'data' in response.keys():
        return response['data']
    else:
        return response

def get_active_devices(client):
    ''' Get active devices '''
    dev_url = '/api/resources/inventory/v1/Device/all'
    devices_data = client.get(dev_url)
    devices = []
    for device in devices_data['data']:
        try:
            if device['result']['value']['streamingStatus'] == "STREAMING_STATUS_ACTIVE":
                devices.append(device['result']['value']['hostname'])
        # pass on archived datasets
        except KeyError as e:
            continue
    return devices

def get_all_device_tags(client):
    tag_url = '/api/resources/tag/v1/DeviceTag/all'
    tag_data = client.get(tag_url)
    tags = []
    for tag in tag_data['data']:
        tags.append({tag['result']['value']['key']['label']:tag['result']['value']['key']['value']})
    return tags

def get_all_interface_tags(client):
    tag_url = '/api/resources/tag/v1/InterfaceTagAssignmentConfig/all'
    tags = client.get(tag_url)
    return tags['data']

def filter_interface_tag(client, dId=None, ifId=None, label=None, value=None):
    tag_url = '/api/resources/tag/v1/InterfaceTagAssignmentConfig/all'
    payload = {
                "partialEqFilter": [
                    {"key": {"deviceId": dId, "interfaceId": ifId, "label": label, "value": value}}
                ]
            }
    response = client.post(tag_url, data=payload)
    return response

def create_itag(client, label, value):
    tag_url = '/api/resources/tag/v1/InterfaceTagConfig'
    payload = {"key":{"label":label,"value":value}}
    response = client.post(tag_url, data=payload)
    return response

def assign_itag(client, dId, ifId, label, value):
    tag_url = '/api/resources/tag/v1/InterfaceTagAssignmentConfig'
    payload = {"key":{"label":label, "value":value, "deviceId": dId, "interfaceId": ifId}}
    response = client.post(tag_url, data=payload)
    return response

def create_dtag(client, label, value):
    tag_url = '/api/resources/tag/v1/DeviceTagConfig'
    payload = {"key":{"label":label,"value":value}}
    response = client.post(tag_url, data=payload)
    return response

def assign_dtag(client, dId, label, value):
    tag_url = '/api/resources/tag/v1/DeviceTagAssignmentConfig'
    payload = {"key":{"label":label, "value":value, "deviceId": dId}}
    response = client.post(tag_url, data=payload)
    return response

### Uncomment the below functions/print statement to test

# ### Get all active events
# print ('=== All active events ===')
# cvpevents = get_events_all(clnt)
# for event in cvpevents:
#     print(event)

# ### Get a specific event
# key = "6098ae39e4c8a9d7"
# ts ="2021-04-06T21:53:00Z"
# get_event(clnt, key, ts)

# ### Get events between two dates
# t1 = "2021-04-06T09:00:00Z"
# t2 = "2021-04-06T14:00:00Z"
# events = get_events_t1_t2(clnt, t1, t2)
# print(f"=== Events between {t1} and {t2} ===")
# pp(events)

# ### Get all INFO severity events ###
# # EVENT_SEVERITY_UNSPECIFIED = 0
# # EVENT_SEVERITY_INFO =	1
# # EVENT_SEVERITY_WARNING = 2
# # EVENT_SEVERITY_ERROR = 3
# # EVENT_SEVERITY_CRITICAL =	4
# ####################################

# severity = 1 ## Severity INFO
# info = get_events_by_severity(clnt, severity)
# print('=== Get all INFO severity events ===')
# pp(info)

# ### Get specific event types

# etype = "LOW_DEVICE_DISK_SPACE"
# event = get_events_by_type(clnt, etype)
# print('=== Get all Low Disk Space events ===')
# pp(event)

# ### Get the inventory
# print ('=== Inventory ===')
# print(get_active_devices(clnt))

# ### Get all devie tags
# print('=== Device Tags ===' )
# for tag in get_all_device_tags(clnt):
#     print (tag)

# ### Get all interface tag assignments
# print(get_all_interface_tags(clnt))

# ### Get all interfaces that have a tag with a specific value on a device
# print(filter_interface_tag(clnt, dId="JPE14070534", value="speed40Gbps"))

# ### Get all tags for an interface of a device
# print(filter_interface_tag(clnt, dId="JPE14070534", ifId="Ethernet1"))

# ### Get all interfaces that have a specific tag assigned
# print(filter_interface_tag(clnt, dId="JPE14070534", label="lldp_hostname"))

# ### Create an interface tag
# create_itag(clnt, "lldp_chassis", "50:08:00:0d:00:48")

# ### Assign an interface tag
# assign_itag(clnt, "JPE14070534", "Ethernet4", "lldp_chassis", "50:08:00:0d:00:38")

# ### Create a device tag
# create_dtag(clnt, "topology_hint_pod", "ire-pod11")

# ### Assign an interface tag
# assign_dtag(clnt, "JPE14070534", "topology_hint_pod", "ire-pod11" )



# Copyright (c) 2021 Arista Networks, Inc.
# Use of this source code is governed by the Apache License 2.0
# that can be found in the COPYING file.
#
# Get configlets and save them to individual files using multi threading
from cvprac.cvp_client import CvpClient
import ssl
ssl._create_default_https_context = ssl._create_unverified_context
import requests.packages.urllib3
requests.packages.urllib3.disable_warnings()
from concurrent.futures import ThreadPoolExecutor
from functools import wraps

# Create connection to CloudVision
clnt = CvpClient()
clnt.connect(['cvp1'],'username', 'password')

total = clnt.api.get_configlets(start=0,end=1)['total']

def get_list_of_configlets():
    """
    Create a thread pool and download specified urls
    """

    futures_list = []
    results = []

    with ThreadPoolExecutor(max_workers=40) as executor:
        for i in range(0,total+1,10):
            futures = executor.submit(clnt.api.get_configlets, start=i,end=i+10)
            futures_list.append(futures)

        for future in futures_list:
            try:
                result = future.result(timeout=60)
                results.append(result)
            except Exception:
                results.append(None)
                print(future.result())
    return results

if __name__ == "__main__":

    results = get_list_of_configlets()
    for future in results:
        for configlet in future['data']:
            with open(configlet['name'],'w') as f:
                f.write(configlet['config'])



# Copyright (c) 2021 Arista Networks, Inc.
# Use of this source code is governed by the Apache License 2.0
# that can be found in the COPYING file.

from cvprac.cvp_client import CvpClient
import json
import ssl
ssl._create_default_https_context = ssl._create_unverified_context
import requests.packages.urllib3
requests.packages.urllib3.disable_warnings()

# Create connection to CloudVision
clnt = CvpClient()
clnt.connect(['cvp1'],'username', 'password')

image_name = "vEOS-4.26.0.1F"
image = clnt.api.get_image_bundle_by_name(image_name)

device_name = "tp-avd-leaf2"
device  = clnt.api.get_device_by_name(device_name)

def apply_image_to_element_no_temp(image, element, name, id_type, create_task=True):
    ''' Apply an image bundle to a device or container
        A copy of the appl_image_to_element() function without creating a tempAction.
        Useful in situations where we need to call saveTopology on a per tempAction basis,
        which is only possible if the addTempAction function is not used and the data
        that we would've passed in the addTempAction call is passed in the
        saveTopology call.
        Args:
            image (dict): The image info.
            element (dict): Info about the element to apply an image to. Dict
                can contain device info or container info.
            name (str): Name of the element the image is being applied to.
            id_type (str): - Id type of the element the image is being applied to
                            - can be 'netelement' or 'container'
            create_task (bool): Determines whether or not to execute a save
                and create the tasks (if any)
        Returns:
            response (list): A list that contains the tempAction data
                Ex: [{'NetworkRollbackTask': False,
                      'taskJson': '[{
                        "info": "Apply image: vEOS-4.26.0.1F to netelement tp-avd-leaf2",
                        "infoPreview": "Apply image: vEOS-4.26.0.1F to netelement tp-avd-leaf2",
                        "note": "",
                        "action": "associate", "nodeType":
                        "imagebundle",
                        "nodeId": "imagebundle_1622072231719691917",
                        "toId": "50:08:00:b1:5b:0b",
                        "toIdType": "netelement",
                        "fromId": "",
                        "nodeName": "vEOS-4.26.0.1F",
                        "fromName": "", "
                        toName": "tp-avd-leaf2",
                        "childTasks": [],
                        "parentTask": ""}]'}]
    '''

    print('Attempt to apply %s to %s %s' % (image['name'],
                                                        id_type, name))
    info = 'Apply image: %s to %s %s' % (image['name'], id_type, name)
    node_id = ''
    if 'imageBundleKeys' in image:
        if image['imageBundleKeys']:
            node_id = image['imageBundleKeys'][0]
        print('Provided image is an image object.'
                        ' Using first value from imageBundleKeys - %s'
                        % node_id)
    if 'id' in image:
        node_id = image['id']
        print('Provided image is an image bundle object.'
                        ' Found v1 API id field - %s' % node_id)
    elif 'key' in image:
        node_id = image['key']
        print('Provided image is an image bundle object.'
                        ' Found v2 API key field - %s' % node_id)
    data = [
      {
      "NetworkRollbackTask": False,
      "taskJson": json.dumps([{'info': info,
                        'infoPreview': info,
                        'note': '',
                        'action': 'associate',
                        'nodeType': 'imagebundle',
                        'nodeId': node_id,
                        'toId': element['key'],
                        'toIdType': id_type,
                        'fromId': '',
                        'nodeName': image['name'],
                        'fromName': '',
                        'toName': name,
                        'childTasks': [],
                        'parentTask': ''}])
      }
    ]
    return data

create_task = False
tempAction = apply_image_to_element_no_temp(image, device, device['fqdn'], 'netelement', create_task)

clnt.api._save_topology_v2(tempAction)


# Copyright (c) 2021 Arista Networks, Inc.
# Use of this source code is governed by the Apache License 2.0
# that can be found in the COPYING file.

from cvprac.cvp_client import CvpClient
import ssl
ssl._create_default_https_context = ssl._create_unverified_context
import requests.packages.urllib3
requests.packages.urllib3.disable_warnings()

# Create connection to CloudVision
clnt = CvpClient()
clnt.connect(['cvp1'],'username', 'password')


data = {"data":[{"info":"Device IP Address Changed",
                "infoPreview":"<b> Device IP Address Changed to 10.83.13.214</b>",
                "action":"associate",
                "nodeType":"ipaddress",
                "nodeId":"",
                "toId":"50:08:00:a7:ca:c3",          # MAC of the device
                "fromId":"",
                "nodeName":"",
                "fromName":"",
                "toName":"tp-avd-leaf1",             # hostname
                "toIdType":"netelement",
                "nodeIpAddress":"10.83.13.219",      # the temporary management IP Address
                "nodeTargetIpAddress":"10.83.13.214" # the final management IP address
                }
            ]
        }
clnt.api._add_temp_action(data)

clnt.api._save_topology_v2([])



# Copyright (c) 2021 Arista Networks, Inc.
# Use of this source code is governed by the Apache License 2.0
# that can be found in the COPYING file.

# In this example we are going to assign topology tags using the tags.v2 workspace aware API
# More details on tag.v2 can be found at https://aristanetworks.github.io/cloudvision-apis/models/tag.v2/
# NOTE: Tag.v2 can be used for assigning both device and interface tags (studios, topology, etc) and it's not
# limited to topology tags only. 
# The following are some of the built-in tags that can be used to modify the Topology rendering:
# topology_hint_type: < core | edge | endpoint | management | leaf | spine  >
# topology_hint_rack: < rack name as string >
# topology_hint_pod: < pod name as string >
# topology_hint_datacenter: < datacenter name as string >
# topology_hint_building: < building name as string >
# topology_hint_floor: < floor name as string >
# topology_network_type: < datacenter | campus | cloud >

from cvprac.cvp_client import CvpClient
import uuid
from datetime import datetime

# Reading the service account token from a file
with open("token.tok") as f:
    token = f.read().strip('\n')

# Create connection to CloudVision
clnt = CvpClient()
clnt.connect(nodes=['cvp1'], username='',password='',api_token=token)

tags_common = [{"key": "topology_hint_pod", "value": "tp-avd-pod1"},
               {"key": "topology_hint_datacenter", "value": "tp-avd-dc1"}]
tags_leaf1 = [{"key": "topology_hint_rack", "value": "tp-avd-leafs1"},
              {"key": "topology_hint_type", "value": "leaf"}]
tags_leaf2 = [{"key": "topology_hint_rack", "value": "tp-avd-leafs2"},
              {"key": "topology_hint_type", "value": "leaf"}]
tags_spines = [{"key": "topology_hint_rack", "value": "tp-avd-spines"},
               {"key": "topology_hint_type", "value": "spine"}]

# Create workspace
display_name = f"Change_{datetime.now().strftime('%Y%m%d_%H%M%S')}"
workspace_id = str(uuid.uuid4())
clnt.api.workspace_config(workspace_id,display_name)

### Create tags
element_type = "ELEMENT_TYPE_DEVICE"

for tag in tags_common+tags_leaf1+tags_leaf2+tags_spines:
    tag_label = tag['key']
    tag_value = tag['value']
    clnt.api.tag_config(element_type, workspace_id, tag_label, tag_value)

### Assign tags
devices = {"leafs1":["BAD032986065E8DC14CBB6472EC314A6","0123F2E4462997EB155B7C50EC148767"],
		   "leafs2":["8520AF39790A4EC959550166DC5DEADE", "6323DA7D2B542B5D09630F87351BEA41"],
		   "spines":["CD0EADBEEA126915EA78E0FB4DC776CA", "2568DB4A33177968A78C4FD5A8232159"]}

for tag in tags_common+tags_leaf1:
    tag_label = tag['key']
    tag_value = tag['value']
    interface_id = ''
    for leaf in devices['leafs1']:
        device_id = leaf
        clnt.api.tag_assignment_config(element_type, workspace_id, tag_label, tag_value, device_id, interface_id)
for tag in tags_common+tags_leaf2:
    tag_label = tag['key']
    tag_value = tag['value']
    interface_id = ''
    for leaf in devices['leafs2']:
        device_id = leaf
        clnt.api.tag_assignment_config(element_type, workspace_id, tag_label, tag_value, device_id, interface_id)
for tag in tags_common+tags_spines:
    tag_label = tag['key']
    tag_value = tag['value']
    interface_id = ''
    for spine in devices['spines']:
        device_id = spine
        clnt.api.tag_assignment_config(element_type, workspace_id, tag_label, tag_value, device_id, interface_id)

### Start build
request = 'REQUEST_START_BUILD'
request_id = 'b1'
description='testing cvprac build'
clnt.api.workspace_config(workspace_id=workspace_id, display_name=display_name, 
                                     description=description, request=request, request_id=request_id)

### Check workspace build status and proceed only after it finishes building
b = 0
while b == 0:
    build_id = request_id
    # Requesting for the build status too fast might fail if the build start didn't finish creating
    # the build with the request_id/build_id
    while True:
        try:
            request = clnt.api.workspace_build_status(workspace_id, build_id)
            break
        except Exception as e:
            continue
    if request['value']['state'] == 'BUILD_STATE_SUCCESS':
        b = b+1
    else:
        continue

### Submit workspace
request = 'REQUEST_SUBMIT'
request_id = 's1'
clnt.api.workspace_config(workspace_id=workspace_id,display_name=display_name,description=description,request=request,request_id=request_id)



# Copyright (c) 2021 Arista Networks, Inc.
# Use of this source code is governed by the Apache License 2.0
# that can be found in the COPYING file.

from cvprac.cvp_client import CvpClient
import ssl
ssl._create_default_https_context = ssl._create_unverified_context
import requests.packages.urllib3
requests.packages.urllib3.disable_warnings()

# Create connection to CloudVision
clnt = CvpClient()
clnt.connect(nodes=['cvp1'], username="username",password="password")

# Get the full inventory
inventory = clnt.api.get_inventory()

# Create a list of MAC addresses
device_macs = []
for i in inventory:
    device_macs.append(i['systemMacAddress'])

# Create a dictionary with MAC to running-config mapping
running_configs = {}
for i in device_macs:
    running_configs[i] = clnt.api.get_device_configuration(i)

# Write the running-configs of each device using the hostname as the filename
for i in inventory:
    with open(i['fqdn']+'.cfg', 'w') as f:
        f.write(running_configs[i['systemMacAddress']])


#
# Copyright (c) 2017, Arista Networks, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
#
#   Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
#
#   Neither the name of Arista Networks nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# 'AS IS' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ARISTA NETWORKS
# BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
# IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

''' CVP Restful API client exception classes
'''

class CvpClientError(Exception):
    ''' CVP Restful API client error
    '''
    def __init__(self, msg):
        Exception.__init__(self)
        self.msg = msg

    def __str__(self):
        return self.msg

class CvpApiError(CvpClientError):
    ''' Error encountered related to the CVP API request.
    '''
    def __init__(self, msg):
        CvpClientError.__init__(self, msg)

class CvpLoginError(CvpClientError):
    ''' Error encountered trying to login into CVP.
    '''
    def __init__(self, msg):
        CvpClientError.__init__(self, msg)

class CvpRequestError(CvpClientError):
    ''' CVP request not properly constructed.
    '''
    def __init__(self, msg):
        CvpClientError.__init__(self, msg)

class CvpSessionLogOutError(CvpClientError):
    ''' Current CVP session has been logged out.
    '''
    def __init__(self, msg):
        CvpClientError.__init__(self, msg)



#
# Copyright (c) 2017, Arista Networks, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
#
#   Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
#
#   Neither the name of Arista Networks nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ARISTA NETWORKS
# BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
# IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

# pylint: disable=too-many-branches,too-many-statements,too-many-locals,too-many-lines

''' RESTful API Client class for Cloudvision(R) Portal

This module provides a RESTful API client for Cloudvision(R) Portal (CVP)
which can be used for building applications that work with Arista CVP.

When the class is instantiated the logging is configured.  Either syslog,
file logging, both, or none can be enabled.  If neither syslog nor filename is
specified then no logging will be performed.

This class supports creating a connection to a CVP node and then issuing
subsequent GET and POST requests to CVP.  A GET or POST request will be
automatically retried on the same node if the request receives a
requests.exceptions.Timeout or ReadTimeout error.  A GET or POST request will
be automatically retried on the same node if the request receives a
CvpSessionLogOutError.  For this case a login will be performed before the
request is retried.  For either case, the maximum number of times a request
will be retried on the same node is specified by the class attribute
NUM_RETRY_REQUESTS.

If more than one CVP node is specified when creating a connection, and a GET
or POST request that receives a requests.exceptions.ConnectionError,
requests.exceptions.HTTPError, or a requests.exceptions.TooManyRedirects will
be retried on the next CVP node in the list.  If a GET or POST request that
receives a requests.exceptions.Timeout or CvpSessionLogOutError and the retries
on the same node exceed NUM_RETRY_REQUESTS, then the request will be retried
on the next node on the list.

If any of the errors persists across all nodes then the GET or POST request
will fail and the last error that occurred will be raised.

The class provides connect, get, and post methods that allow the user to make
direct RESTful API calls to CVP.

Example:

    >>> from cvprac.cvp_client import CvpClient
    >>> clnt = CvpClient()
    >>> clnt.connect(['cvp1', 'cvp2', 'cvp3'], 'cvp_user', 'cvp_word')
    >>> result = clnt.get('/cvpInfo/getCvpInfo.do')
    >>> print result
    {u'version': u'2016.1.0'}
    >>>

The class provides a wrapper function around the CVP RESTful API operations.
Each API method takes the RESTful API parameters as method parameters to the
operation method.  The API class was added to the client class because the
API functions are required when using the CVP RESTful API and placing them
in this library avoids duplicating the calls in every application that uses
this class.

Example:

    >>> from cvprac.cvp_client import CvpClient
    >>> clnt = CvpClient()
    >>> clnt.connect(['cvp1', 'cvp2', 'cvp3'], 'cvp_user', 'cvp_word')
    >>> result = clnt.api.get_cvp_info()
    >>> print result
    {u'version': u'2016.1.0'}
    >>>
'''

import os
import re
import json
import logging
from logging.handlers import SysLogHandler
from itertools import cycle
from packaging.version import parse

import requests
from requests.exceptions import ( # pylint: disable=redefined-builtin
    ConnectionError,
    HTTPError,
    Timeout,
    ReadTimeout,
    TooManyRedirects,
    JSONDecodeError
)

from cvprac.cvp_api import CvpApi
from cvprac.cvp_client_errors import CvpApiError, CvpLoginError, \
    CvpRequestError, CvpSessionLogOutError


class CvpClient():
    ''' Use this class to create a persistent connection to CVP.
    '''
    # pylint: disable=too-many-instance-attributes
    # Maximum number of times to retry a get or post to the same
    # CVP node.
    NUM_RETRY_REQUESTS = 3
    LATEST_API_VERSION = 9.0

    def __init__(self, logger='cvprac', syslog=False, filename=None,
                 log_level='INFO'):
        ''' Initialize the client and configure logging.  Either syslog, file
            logging, both, or none can be enabled.  If neither syslog
            nor filename is specified then no logging will be performed.

            Args:
                logger (str): The name assigned to the logger.
                syslog (bool): If True enable logging to syslog. Default is
                    False.
                filename (str): Log to the file specified by filename. Default
                    is None.
                log_level (str): Log level to use for logger. Default is INFO.
        '''
        self.apiversion = None
        self.authdata = None
        self.cert = False
        self.connect_timeout = None
        self.cookies = None
        self.error_msg = ''
        self.node_cnt = None
        self.node_pool = None
        self.nodes = None
        self.port = None
        self.protocol = None
        self.session = None
        self.url_prefix = None
        self.url_prefix_short = None
        self.is_cvaas = False
        self.tenant = None
        self.cvaas_token = None
        self.api_token = None
        self.version = None
        self._last_used_node = None
        self.proxies = None

        # Save proper headers
        self.headers = {'Accept': 'application/json',
                        'Content-Type': 'application/json'}

        self.log = logging.getLogger(logger)
        self.set_log_level(log_level)
        if syslog:
            # Enables sending logging messages to the local syslog server.
            self.log.addHandler(SysLogHandler())
        if filename:
            # Enables sending logging messages to a file.
            self.log.addHandler(logging.FileHandler(filename))
        if syslog is False and filename is None:
            # Not logging so use the null handler
            self.log.addHandler(logging.NullHandler())

        # Instantiate the CvpApi class
        self.api = CvpApi(self)

    @property
    def last_used_node(self):
        ''' Returns the node that the last request was sent to regardless of
            whether the request was successful or not.

            Returns:
                String identifying the node that the last request was sent to.
        '''
        return self._last_used_node

    def set_log_level(self, log_level='INFO'):
        ''' Set log level for logger. Defaults to INFO if no level passed in or
            if an invalid level is passed in.

            Args:
                log_level (str): Log level to use for logger. Default is INFO.
        '''
        log_level = log_level.upper()
        if log_level not in ['NOTSET', 'DEBUG', 'INFO',
                             'WARNING', 'ERROR', 'CRITICAL']:
            log_level = 'INFO'
        self.log.setLevel(getattr(logging, log_level))

    def set_version(self, version):
        ''' Set the CVP API version to be used when making api calls.

            For CVP versions 2018.1.X and prior, use api version 1.0
            For CVP versions 2018.2.X, use api version 2.0
            For CVP versions 2019.0.0 through 2020.1.0, use api version 3.0
            For CVP versions 2020.1.1 through 2020.2.3, use api version 4.0
            For CVP versions 2020.2.4 through 2021.1.x, use api version 5.0
            For CVP versions 2021.2.x, use api version 6.0
            For CVP versions 2021.3.x, use api version 7.0
            For CVP versions 2022.1.0 and beyond, use api version 8.0

            Args:
                version (str): The CVP version in use.
        '''
        self.version = version
        self.log.info('Version %s', version)
        # Set apiversion to latest available API version for CVaaS
        # Set apiversion to 9.0 for 2023.1.x
        # Set apiversion to 8.0 for 2022.1.x - 2022.3.x
        # Set apiversion to 7.0 for 2021.3.x
        # Set apiversion to 6.0 for 2021.2.x
        # Set apiversion to 5.0 for 2020.2.4 through 2021.1.x
        # Set apiversion to 4.0 for 2020.1.1 through 2020.2.3
        # Set apiversion to 3.0 for 2019.0.0 through 2020.1.0
        # Set apiversion to 2.0 for 2018.2.X
        # Set apiversion to 1.0 for 2018.1.X and prior
        if self.is_cvaas:
            self.log.info('Setting API version to %d for CVaaS',
                          self.LATEST_API_VERSION)
            self.apiversion = self.LATEST_API_VERSION
        else:
            version_components = version.split(".")
            if len(version_components) < 3:
                version_components.append("0")
                self.log.info('Version found with less than 3 components.'
                              ' Appending 0. Updated Version String - %s',
                              ".".join(version_components))
            full_version = ".".join(version_components)
            if parse(full_version) >= parse('2024.1.0'):
                self.log.info('Setting API version to v12')
                self.apiversion = 12.0
            elif parse(full_version) >= parse('2023.3.0'):
                self.log.info('Setting API version to v11')
                self.apiversion = 11.0
            elif parse(full_version) >= parse('2023.2.0'):
                self.log.info('Setting API version to v10')
                self.apiversion = 10.0
            elif parse(full_version) >= parse('2023.1.0'):
                self.log.info('Setting API version to v9')
                self.apiversion = 9.0
            elif parse(full_version) >= parse('2022.1.0'):
                self.log.info('Setting API version to v8')
                self.apiversion = 8.0
            elif parse(full_version) >= parse('2021.3.0'):
                self.log.info('Setting API version to v7')
                self.apiversion = 7.0
            elif parse(full_version) >= parse('2021.2.0'):
                self.log.info('Setting API version to v6')
                self.apiversion = 6.0
            elif parse(full_version) >= parse('2020.2.4'):
                self.log.info('Setting API version to v5')
                self.apiversion = 5.0
            elif parse(full_version) >= parse('2020.1.1'):
                self.log.info('Setting API version to v4')
                self.apiversion = 4.0
            elif parse(full_version) >= parse('2019.0.0'):
                self.log.info('Setting API version to v3')
                self.apiversion = 3.0
            elif parse(full_version) >= parse('2018.2.0'):
                self.log.info('Setting API version to v2')
                self.apiversion = 2.0
            else:
                self.log.info('Setting API version to v1')
                self.apiversion = 1.0

    def connect(self, nodes, username, password, connect_timeout=10,
                request_timeout=30, protocol='https', port=None, cert=False,
                is_cvaas=False, tenant=None, api_token=None, cvaas_token=None,
                proxies=None):
        ''' Login to CVP and get a session ID and cookie.  Currently
            certificates are not verified if the https protocol is specified. A
            warning may be printed out from the requests module for this case.

            Args:
                nodes (list): A list of hostname/IP addresses for CVP nodes
                username (str): The CVP username
                password (str): The CVP password
                connect_timeout (int): The number of seconds to wait for a
                    connection.
                request_timeout (int): The default number of seconds to allow
                    api requests to complete before timing out.
                protocol (str): The protocol to use to connect to CVP.
                    THIS PARAMETER IS NOT USED AND WILL BE DEPRECATED.
                    ONLY INCLUDED TO NOT BREAK EXISTING CODE THAT HAS PROTOCOL
                    SPECIFIED IN CONNECTION.
                port (int): The TCP port of the endpoint for the connection.
                    If this keyword is not specified, the default value is
                    automatically determined by the transport type.
                    (http=80, https=443)
                cert (str or boolean): Path to a cert file used for a https
                    connection or boolean with default False. If a cert is
                    provided then the connection will not attempt to fallback
                    to http. The False default sets the request to not verify
                    the servers TLS certificate.
                is_cvaas (boolean): Flag for enabling connection to CVaaS.
                tenant: (string): Tenant/Org within CVaaS to connect to.
                    Required if is_cvaas is enabled.
                cvaas_token (string): API Token to use in place of UN/PW login
                    for CVaaS.
                api_token (string): API Token to use in place of UN/PW login
                    for CVP 2020.3.0 and beyond.
                proxies (dict): A dictionary of proxy protocol to URL. Example:

                    {'http': 'hostname.domain.com:8080',
                     'https': 'hostname.domain.com:8080'}

                     Proxies can also be set via environment variables.
                     Please reference the below link for details of precedence.
                     https://requests.readthedocs.io/en/latest/user/advanced/#proxies

            Raises:
                CvpLoginError: A CvpLoginError is raised if a connection
                    could not be established to any of the nodes.
                TypeError: A TypeError is raised if the nodes argument is not
                    a list.
                ValueError: A ValueError is raised if a port is not specified
                    and the protocol is not http or https.
        '''
        # pylint: disable=too-many-arguments
        if not isinstance(nodes, list):
            raise TypeError('nodes argument must be a list')

        for idx, _ in enumerate(nodes):
            if (os.environ.get('CURRENT_NODE_IP') and
                    nodes[idx] in ['127.0.0.1', 'localhost']):
                # We set this env in script-executor container.
                # Mask localhost or 127.0.0.1 with node IP if this
                # is called from configlet builder scripts.
                nodes[idx] = os.environ.get('CURRENT_NODE_IP')

        self.cert = cert
        self.nodes = nodes
        self.node_cnt = len(nodes)
        self.node_pool = cycle(nodes)
        self.authdata = {'userId': username, 'password': password}
        self.connect_timeout = connect_timeout
        self.api.request_timeout = request_timeout
        # protocol is deprecated and not used.
        self.protocol = protocol
        self.port = port
        self.is_cvaas = is_cvaas
        self.tenant = tenant
        if cvaas_token is not None:
            self.log.warning('The cvaas_token parameter will be deprecated'
                             ' soon. Please start using the api_token'
                             ' parameter instead. It provides the same'
                             ' functionality that was previously provided'
                             ' by cvaas_token. The api_token parameter is'
                             ' a more general API token parameter because'
                             ' using the CVP REST API via token is also'
                             ' available for on premises CVP as of'
                             ' CVP version 2020.3.0')
            self.cvaas_token = cvaas_token
            self.api_token = cvaas_token
        if api_token is not None:
            self.log.warning('Using the new api_token parameter.'
                             ' This will override usage of the cvaas_token'
                             ' parameter if both are provided. This is because'
                             ' api_token and cvaas_token parameters are for'
                             ' the same use case and api_token is more'
                             ' generic')
            self.api_token = api_token
            self.cvaas_token = api_token
        self.proxies = proxies
        self._create_session(all_nodes=True)
        # Verify that we can connect to at least one node
        if not self.session:
            raise CvpLoginError(self.error_msg)

    def _create_session(self, all_nodes=False):
        ''' Login to CVP and get a session ID and user information.
            If the all_nodes parameter is True then try creating a session
            with each CVP node.  If False, then try creating a session with
            each node except the one currently connected to.
        '''
        num_nodes = self.node_cnt
        if not all_nodes and num_nodes > 1:
            num_nodes -= 1

        self.error_msg = '\n'
        for _ in range(0, num_nodes):
            host = next(self.node_pool)
            self.url_prefix = f"https://{host}:{self.port or 443}/web"
            self.url_prefix_short = f"https://{host}:{self.port or 443}"
            error = self._reset_session()
            if error is None:
                break
            self.error_msg += f"{host}: {error}\n"

    def _reset_session(self):
        ''' Get a new request session and try logging into the current
            CVP node. If the login succeeded None will be returned and
            self.session will be valid. If the login failed then an
            exception error will be returned and self.session will
            be set to None.
        '''
        self.session = requests.Session()
        if self.proxies:
            self.session.proxies.update(self.proxies)
        return_error = None
        try:
            self._login()
        except (ConnectionError, CvpApiError, CvpRequestError,
                CvpSessionLogOutError, HTTPError, ReadTimeout, Timeout,
                TooManyRedirects) as error:
            self.log.error(error)
            # Use outer scope var for return to handle
            # Python 3 UnboundLocalError
            return_error = error
            # Any error that occurs during login is a good reason not to use
            # this CVP node.
            self.session = None
        return return_error

    def _is_good_response(self, response, prefix):
        ''' Check for errors in a response from a GET or POST request.
            The response argument contains a response object from a GET or POST
            request.  The prefix argument contains the prefix to put into the
            error message.

            Raises:
                CvpApiError: A CvpApiError is raised if there was a JSON error.
                CvpRequestError: A CvpRequestError is raised if the request
                    is not properly constructed.
                CvpSessionLogOutError: A CvpSessionLogOutError is raised if
                    response from server indicates session was logged out.
        '''
        if not response.ok:
            if 'Unauthorized' in response.reason:
                # Check for 'Unauthorized' User error because this is how
                # CVP responds to a logged out users requests in 2018.x.
                msg = f"{prefix}: Request Error: {response.reason}"
                self.log.error(msg)
                raise CvpApiError(msg)
            if 'User is unauthorized' in response.text:
                # Check for 'User is unauthorized' response text because this
                # is how CVP responds to a logged out users requests in 2019.x.
                msg = f"{prefix}: Request Error: User is unauthorized"
                self.log.error(msg)
                raise CvpApiError(msg)
            msg = f"{prefix}: Request Error: {response.reason} - {response.text}"
            raise CvpRequestError(msg)

        if 'LOG OUT MESSAGE' in response.text:
            msg = f"{prefix}: Request Error: session logged out"
            raise CvpSessionLogOutError(msg)

        joutput = json_decoder(response.text)
        err_code_val = self._finditem(joutput, 'errorCode')
        if err_code_val:
            if 'errorMessage' in joutput:
                err_msg = joutput['errorMessage']
            else:
                if 'errors' in joutput:
                    error_list = joutput['errors']
                else:
                    error_list = [joutput['errorCode']]
                # Build the error message from all the errors.
                err_msg = error_list[0]
                for idx in range(1, len(error_list)):
                    err_msg = f"{err_msg}\n{error_list[idx]}"

            msg = f"{prefix}: Request Error: {err_msg}"
            self.log.error(msg)
            raise CvpApiError(msg)

    def _check_response_status(self, response, prefix):
        ''' Check for status OK in a response from a GET or POST request.
            The response argument contains a response object from a GET or POST
            request.  The prefix argument contains the prefix to put into the
            error message.

            Raises:
                CvpRequestError: A CvpRequestError is raised if request
                response status is not OK.
        '''
        if not response.ok:
            msg = f"{prefix}: Request Error: {response.reason} - {response.text}"
            self.log.error(msg)
            raise CvpRequestError(msg)

    def _login(self):
        ''' Make a POST request to CVP login authentication.
            An error can be raised from the post method call or the
            _is_good_response method call.  Any errors raised would be a good
            reason not to use this host.

            Raises:
                ConnectionError: A ConnectionError is raised if there was a
                    network problem (e.g. DNS failure, refused connection, etc)
                CvpApiError: A CvpApiError is raised if there was a JSON error.
                CvpRequestError: A CvpRequestError is raised if the request
                    is not properly constructed.
                CvpSessionLogOutError: A CvpSessionLogOutError is raised if
                    response from server indicates session was logged out.
                HTTPError: A HTTPError is raised if there was an invalid HTTP
                    response.
                ReadTimeout: A ReadTimeout is raised if there was a request
                    timeout when reading from the connection.
                Timeout: A Timeout is raised if there was a request timeout.
                TooManyRedirects: A TooManyRedirects is raised if the request
                    exceeds the configured number of maximum redirections
                ValueError: A ValueError is raised when there is no valid
                    CVP session.  This occurs because the previous get or post
                    request failed and no session could be established to a
                    CVP node.  Destroy the class and re-instantiate.
        '''
        # Remove any previous session id from the headers
        self.headers.pop('APP_SESSION_ID', None)
        if self.api_token is not None:
            return self._set_headers_api_token()
        if self.is_cvaas:
            raise CvpLoginError('CVaaS only supports API token authentication.'
                                ' Please create an API token and provide it'
                                ' via the api_token parameter in combination'
                                ' with the is_cvaas parameter')
        return self._login_on_prem()

    def _login_on_prem(self):
        ''' Make a POST request to CVP login authentication.
            An error can be raised from the post method call or the
            _is_good_response method call.  Any errors raised would be a good
            reason not to use this host.

            Raises:
                ConnectionError: A ConnectionError is raised if there was a
                    network problem (e.g. DNS failure, refused connection, etc)
                CvpApiError: A CvpApiError is raised if there was a JSON error.
                CvpRequestError: A CvpRequestError is raised if the request
                    is not properly constructed.
                CvpSessionLogOutError: A CvpSessionLogOutError is raised if
                    response from server indicates session was logged out.
                HTTPError: A HTTPError is raised if there was an invalid HTTP
                    response.
                ReadTimeout: A ReadTimeout is raised if there was a request
                    timeout when reading from the connection.
                Timeout: A Timeout is raised if there was a request timeout.
                TooManyRedirects: A TooManyRedirects is raised if the request
                    exceeds the configured number of maximum redirections
                ValueError: A ValueError is raised when there is no valid
                    CVP session.  This occurs because the previous get or post
                    request failed and no session could be established to a
                    CVP node.  Destroy the class and re-instantiate.
        '''
        url = self.url_prefix + '/login/authenticate.do'
        response = self.session.post(url,
                                     data=json.dumps(self.authdata),
                                     headers=self.headers,
                                     timeout=self.connect_timeout,
                                     verify=self.cert)
        self._is_good_response(response, f"Authenticate: {url}")

        self.cookies = response.cookies
        self.headers['APP_SESSION_ID'] = response.json()['sessionId']

    def _set_headers_api_token(self):
        ''' Sets headers with API token instead of making a call to login API.
        '''
        # If using an API token there is no need to run a Login API.
        # Simply add the token into the headers or cookies
        self.headers['Authorization'] = f"Bearer {self.api_token}"
        # Alternative to adding token to headers it can be added to
        # cookies as shown below.
        # self.cookies = {'access_token': self.api_token}
        url = self.url_prefix_short + '/api/v1/rest/'
        response = self.session.get(
            url,
            cookies=self.cookies,
            headers=self.headers,
            timeout=self.connect_timeout,
            verify=self.cert
        )
        # Verify that the generic request was successful
        self._is_good_response(response, f"Authenticate: {url}")

    def logout(self):
        '''

        :return:
        '''
        response = self.post('/login/logout.do')
        if response['data'] == 'success':
            self.log.info('User logged out.')
            self.session = None
        else:
            err = f"Error trying to logout {response}"
            self.log.error(err)

    def _make_request(self, req_type, url, timeout, data=None,
                      files=None):
        ''' Make a GET, POST or DELETE request to CVP.  If the request call raises a
            timeout or CvpSessionLogOutError then the request will be retried
            on the same CVP node.  Otherwise the request will be tried on the
            next CVP node.

            Args:
                req_type (str): Either 'GET', 'POST' or 'DELETE'.
                url (str): Portion of request URL that comes after the host.
                timeout (int): Number of seconds the client will wait between
                    bytes sent from the server.
                data (dict): Dict of key/value pairs to pass as parameters into
                    the request. Default is None.
                files (dict): Dict of file name to files for upload. Currently
                    only used for adding images to CVP. Default is None.

            Returns:
                The JSON response.

            Raises:
                ConnectionError: A ConnectionError is raised if there was a
                    network problem (e.g. DNS failure, refused connection, etc)
                CvpApiError: A CvpApiError is raised if there was a JSON error.
                CvpRequestError: A CvpRequestError is raised if the request
                    is not properly constructed.
                CvpSessionLogOutError: A CvpSessionLogOutError is raised if
                    response from server indicates session was logged out.
                HTTPError: A HTTPError is raised if there was an invalid HTTP
                    response.
                ReadTimeout: A ReadTimeout is raised if there was a request
                    timeout when reading from the connection.
                Timeout: A Timeout is raised if there was a request timeout.
                TooManyRedirects: A TooManyRedirects is raised if the request
                    exceeds the configured number of maximum re-directions
                ValueError: A ValueError is raised when there is no valid
                    CVP session.  This occurs because the previous get, post
                    or delete request failed and no session could be
                    established to a CVP node.  Destroy the class and
                    re-instantiate.
                JSONDecodeError: A JSONDecodeError is raised when the response
                    content contains invalid JSON. Potentially in the case of
                    Resource APIs that will return Stream JSON format with
                    multiple object or in the case where the response contains
                    incomplete JSON.
        '''
        # pylint: disable=too-many-branches
        # pylint: disable=too-many-statements
        # pylint: disable=too-many-arguments
        # pylint: disable=raising-bad-type
        if not self.session:
            raise ValueError('No valid session to CVP node')
        # Keep note of which node is handling this request.
        self._last_used_node = re.match('http[s]?://(.*):',
                                        self.url_prefix).group(1)
        # Retry the request for the number of nodes.
        response = None
        for node_num in range(self.node_cnt):
            # Set full URL based on current node
            if '/api/' in url or '/cvpservice/' in url:
                full_url = self.url_prefix_short + url
            elif self.is_cvaas:
                # For CVaaS use cvpservice instead of web or api
                full_url = self.url_prefix_short + '/cvpservice' + url
            else:
                full_url = self.url_prefix + url
            try:
                response = self._send_request(req_type, full_url, timeout,
                                              data, files)
            except CvpApiError as error:
                # If this is not an Unauthorized CvpApiError raise the error
                # 'Unauthorized' is for 2018.x
                # 'User is unauthorized' is for 2019.x
                if ('Unauthorized' not in error.msg and
                        'User is unauthorized' not in error.msg):
                    raise error
                # If this is the final CVP node raise error
                if node_num + 1 == self.node_cnt:
                    raise error
                # Create a new session to retry on another CVP node.
                self._create_session()
                # Verify that we can connect to at least one node
                # otherwise raise the last error
                if not self.session:
                    raise error
                continue
            except (ConnectionError, HTTPError, TooManyRedirects, ReadTimeout,
                    Timeout, CvpSessionLogOutError) as error:
                # If this is the final CVP node raise error
                if node_num + 1 == self.node_cnt:
                    raise error
                # Create a new session to retry on another CVP node.
                self._create_session()
                # Verify that we can connect to at least one node
                # otherwise raise the last error
                if not self.session:
                    raise error
                continue
            break

        if not response:
            self.log.debug('Received no response for request %s %s',
                           req_type, url)
            return None

        # Added check for response.content being 'null' because of the
        # service account APIs being a special case /services/ API that
        # returns a null string for no objects instead of an empty string.
        if not response.content or response.content == b'null':
            return {'data': []}

        try:
            resp_data = response.json()
            if (resp_data is not None and 'result' in resp_data and
                    '/resources/' in full_url):
                # Resource APIs use JSON streaming and will return
                # multiple JSON objects during GetAll type API
                # calls. We are wrapping the multiple objects into
                # a key "data" and we also return a dictionary with
                # key "data" as an empty dict for no data. This
                # checks and keeps consistent the "data" key wrapper
                # for a Resource API GetAll that returns a single
                # object.
                return {'data': [resp_data]}
            return resp_data
        except JSONDecodeError as error:
            # Truncate long error messages
            err_str = str(error)
            if len(err_str) > 700:
                err_str = f"{err_str[:300]}[... truncated ...]" \
                          f" {err_str[-300:]}"
            self.log.debug('Error trying to decode request response - %s',
                           err_str)
            if 'Extra data' in str(error):
                self.log.debug('Found multiple objects or NO objects in'
                               ' response data. Attempt to decode')
                decoded_data = json_decoder(response.text)
                return {'data': decoded_data}
            self.log.error("Unknown format for JSONDecodeError - %s", err_str)
            raise error

    def _send_request(self, req_type, full_url, timeout, data=None,
                      files=None):
        ''' Make a GET, POST or DELETE request to CVP.  If the request call
            raises a timeout or CvpSessionLogOutError then the request will be
            retried on the same CVP node.  Otherwise the request will be tried
            on the next CVP node.

            Args:
                req_type (str): Either 'GET', 'POST' or 'DELETE'.
                full_url (str): Portion of request URL that comes after the
                    host.
                timeout (int): Number of seconds the client will wait between
                    bytes sent from the server.
                data (dict): Dict of key/value pairs to pass as parameters into
                    the request. Default is None.
                files (dict): Dict of file name to files for upload. Currently
                    only used for adding images to CVP. Default is None.

            Returns:
                The JSON response.

            Raises:
                ConnectionError: A ConnectionError is raised if there was a
                    network problem (e.g. DNS failure, refused connection, etc)
                CvpApiError: A CvpApiError is raised if there was a JSON error.
                CvpRequestError: A CvpRequestError is raised if the request
                    is not properly constructed.
                CvpSessionLogOutError: A CvpSessionLogOutError is raised if
                    response from server indicates session was logged out.
                HTTPError: A HTTPError is raised if there was an invalid HTTP
                    response.
                ReadTimeout: A ReadTimeout is raised if there was a request
                    timeout when reading from the connection.
                Timeout: A Timeout is raised if there was a request timeout.
                TooManyRedirects: A TooManyRedirects is raised if the request
                    exceeds the configured number of maximum re-directions
                ValueError: A ValueError is raised when there is no valid
                    CVP session.  This occurs because the previous get, post
                    or delete request failed and no session could be
                    established to a CVP node.  Destroy the class and
                    re-instantiate.
        '''
        # pylint: disable=too-many-branches
        # pylint: disable=too-many-statements
        # pylint: disable=too-many-arguments
        # pylint: disable=raising-bad-type
        # For get or post requests apply both the connect and read timeout.
        timeout = (self.connect_timeout, timeout)
        for req_try in range(self.NUM_RETRY_REQUESTS):
            try:
                if req_type == 'GET':
                    response = self.session.get(full_url,
                                                cookies=self.cookies,
                                                headers=self.headers,
                                                timeout=timeout,
                                                verify=self.cert)
                elif req_type == 'POST':
                    if files is None:
                        response = self.session.post(full_url,
                                                     cookies=self.cookies,
                                                     data=json.dumps(data),
                                                     headers=self.headers,
                                                     timeout=timeout,
                                                     verify=self.cert)
                    else:
                        fhs = {}
                        fhs['Accept'] = self.headers['Accept']
                        if 'APP_SESSION_ID' in self.headers:
                            fhs['APP_SESSION_ID'] = self.headers[
                                'APP_SESSION_ID']
                        if 'Authorization' in self.headers:
                            fhs['Authorization'] = self.headers[
                                'Authorization']
                        response = self.session.post(full_url,
                                                     cookies=self.cookies,
                                                     headers=fhs,
                                                     timeout=timeout,
                                                     verify=self.cert,
                                                     files=files)
                elif req_type == 'DELETE':
                    response = self.session.delete(full_url,
                                                   cookies=self.cookies,
                                                   data=json.dumps(data),
                                                   headers=self.headers,
                                                   timeout=timeout,
                                                   verify=self.cert)
            except (ConnectionError, HTTPError, TooManyRedirects) as error:
                # Any of these errors is a good reason to try another CVP node
                self.log.error(error)
                raise error
            except (ReadTimeout, Timeout) as error:
                self.log.debug(error)
                # If there was a timeout and this is not the final try,
                # retry this request to the same node. If this is the final
                # try raise the error so another CVP node can be tried
                if req_try + 1 == self.NUM_RETRY_REQUESTS:
                    raise error
                continue

            try:
                self._is_good_response(response, f"{req_type}: {full_url} ")
            except CvpSessionLogOutError as error:
                self.log.debug(error)
                # Retry the request to the same node if there was a CVP session
                # logout. Reset the session which will login. If a valid
                # session comes back then clear the error so this request will
                # be retried on the same node.
                if req_try + 1 == self.NUM_RETRY_REQUESTS:
                    raise error
                self._reset_session()
                if not self.session:
                    raise error
                continue
            except CvpApiError as error:
                self.log.debug(error)
                if ('Unauthorized' in error.msg or
                        'User is unauthorized' in error.msg):
                    # Retry the request to the same node if there was an
                    # Unauthorized User error because this is how CVP responds
                    # to a logged out users requests in 2017.1.
                    # Check for 'User is unauthorized' in error because this is
                    # how CVP responds to a logged out user requests in 2019.x.
                    # Reset the session which will login. If a valid
                    # session comes back then clear the error so this request
                    # will be retried on the same node.
                    if req_try + 1 == self.NUM_RETRY_REQUESTS:
                        raise error
                    self._reset_session()
                    if not self.session:
                        raise error
                    continue
                # pylint: disable=raising-bad-type
                raise error
            return response

    def get(self, url, timeout=30):
        ''' Make a GET request to CVP.  If the request call raises an error
            or if the JSON response contains a CVP session related error then
            retry the request on another CVP node.

            Args:
                url (str): Portion of request URL that comes after the host.
                timeout (int): Number of seconds the client will wait between
                    bytes sent from the server.  Default value is 30 seconds.

            Returns:
                The JSON response.

            Raises:
                ConnectionError: A ConnectionError is raised if there was a
                    network problem (e.g. DNS failure, refused connection, etc)
                CvpApiError: A CvpApiError is raised if there was a JSON error.
                CvpRequestError: A CvpRequestError is raised if the request
                    is not properly constructed.
                CvpSessionLogOutError: A CvpSessionLogOutError is raised if
                    response from server indicates session was logged out.
                HTTPError: A HTTPError is raised if there was an invalid HTTP
                    response.
                ReadTimeout: A ReadTimeout is raised if there was a request
                    timeout when reading from the connection.
                Timeout: A Timeout is raised if there was a request timeout.
                TooManyRedirects: A TooManyRedirects is raised if the request
                    exceeds the configured number of maximum re-directions
                ValueError: A ValueError is raised when there is no valid
                    CVP session.  This occurs because the previous get, post
                    or delete request failed and no session could be
                    established to a CVP node.  Destroy the class and
                    re-instantiate.
        '''
        return self._make_request('GET', url, timeout)

    def post(self, url, data=None, files=None, timeout=30):
        ''' Make a POST request to CVP.  If the request call raises an error
            or if the JSON response contains a CVP session related error then
            retry the request on another CVP node.

            Args:
                url (str): Portion of request URL that comes after the host.
                data (dict): Dict of key/value pairs to pass as parameters into
                    the request. Default is None.
                files (dict): Dict of file name to files for upload. Currently
                    only used for adding images to CVP. Default is None.
                timeout (int): Number of seconds the client will wait between
                    bytes sent from the server.  Default value is 30 seconds.

            Returns:
                The JSON response.

            Raises:
                ConnectionError: A ConnectionError is raised if there was a
                    network problem (e.g. DNS failure, refused connection, etc)
                CvpApiError: A CvpApiError is raised if there was a JSON error.
                CvpRequestError: A CvpRequestError is raised if the request
                    is not properly constructed.
                CvpSessionLogOutError: A CvpSessionLogOutError is raised if
                    response from server indicates session was logged out.
                HTTPError: A HTTPError is raised if there was an invalid HTTP
                    response.
                ReadTimeout: A ReadTimeout is raised if there was a request
                    timeout when reading from the connection.
                Timeout: A Timeout is raised if there was a request timeout.
                TooManyRedirects: A TooManyRedirects is raised if the request
                    exceeds the configured number of maximum re-directions
                ValueError: A ValueError is raised when there is no valid
                    CVP session.  This occurs because the previous get, post
                    or delete request failed and no session could be
                    established to a CVP node.  Destroy the class and
                    re-instantiate.
        '''
        return self._make_request('POST', url, timeout, data=data, files=files)

    def delete(self, url, data=None, timeout=30):
        ''' Make a DELETE request to CVP.  If the request call raises an error
            or if the JSON response contains a CVP session related error then
            retry the request on another CVP node.

            Args:
                url (str): Portion of request URL that comes after the host.
                data (dict): Dict of key/value pairs to pass as parameters into
                    the request. Default is None.
                timeout (int): Number of seconds the client will wait between
                    bytes sent from the server.  Default value is 30 seconds.

            Returns:
                The JSON response.

            Raises:
                ConnectionError: A ConnectionError is raised if there was a
                    network problem (e.g. DNS failure, refused connection, etc)
                CvpApiError: A CvpApiError is raised if there was a JSON error.
                CvpRequestError: A CvpRequestError is raised if the request
                    is not properly constructed.
                CvpSessionLogOutError: A CvpSessionLogOutError is raised if
                    response from server indicates session was logged out.
                HTTPError: A HTTPError is raised if there was an invalid HTTP
                    response.
                ReadTimeout: A ReadTimeout is raised if there was a request
                    timeout when reading from the connection.
                Timeout: A Timeout is raised if there was a request timeout.
                TooManyRedirects: A TooManyRedirects is raised if the request
                    exceeds the configured number of maximum re-directions
                ValueError: A ValueError is raised when there is no valid
                    CVP session.  This occurs because the previous get, post
                    or delete request failed and no session could be
                    established to a CVP node.  Destroy the class and
                    re-instantiate.
        '''
        return self._make_request('DELETE', url, timeout, data=data)

    def _finditem(self, obj, key):
        """ Find a key in a a nested list/dict.

            Args:
                obj (dict): Object to iterate to return value for provided key
                key (str): The key to locate in dict and return the value for

            Returns:
                Value of found key or None if not found.
        """
        item = None
        if isinstance(obj, dict):
            if key in obj:
                item = obj[key]
            else:
                for _, value in obj.items():
                    if isinstance(value, (dict, list)):
                        item = self._finditem(value, key)
                        if item is not None:
                            break
        elif isinstance(obj, list):
            for i in obj:
                if isinstance(i, (dict, list)):
                    item = self._finditem(i, key)
                    if item is not None:
                        break
        return item


def json_decoder(data):
    ''' Check for ...
    '''
    decoder = json.JSONDecoder()
    position = 0
    decoded_data = []
    while True:
        try:
            obj, position = decoder.raw_decode(data, position)
            decoded_data.append(obj)
            position += 1
        except ValueError:
            break
    if len(decoded_data) == 1:
        return decoded_data[0]
    return decoded_data



# Copyright (c) 2021 Arista Networks, Inc.
# Use of this source code is governed by the Apache License 2.0
# that can be found in the COPYING file.

from cvprac.cvp_client import CvpClient
from cvprac.cvp_client_errors import CvpApiError
import ssl
ssl._create_default_https_context = ssl._create_unverified_context
import requests.packages.urllib3
requests.packages.urllib3.disable_warnings()
from getpass import getpass

# Create connection to CloudVision
clnt = CvpClient()
clnt.connect(['cvp1'],'username', 'password')

username = "cvpuser2"
password = getpass()
role = "network-admin"
status = "Enabled"
first_name = "Cloud"
last_name = "Vision"
email = "cvp@arista.com"
utype = "TACACS"

try:
    clnt.api.add_user(username,password,role,status,first_name,last_name,email,utype)
except CvpApiError as e:
    print(e)



# Copyright (c) 2022 Arista Networks, Inc.
# Use of this source code is governed by the Apache License 2.0
# that can be found in the COPYING file.

from cvprac.cvp_client import CvpClient
import ssl
import uuid
import time
ssl._create_default_https_context = ssl._create_unverified_context
import requests.packages.urllib3
requests.packages.urllib3.disable_warnings()

# Create connection to CloudVision
clnt = CvpClient()
clnt.connect(nodes=['cvp1'], username="username", password="password")

device_id = input("Serial number of the device to be decommissioned: ")
request_id = str(uuid.uuid4())
clnt.api.device_decommissioning(device_id, request_id)

# This API call will fully decommission the device, ie remove it from both
# Network Provisioning and Device Inventory (telemetry). It send an eAPI request
# to EOS to shutdown the TerminAttr daemon, waits for streaming to stop and then removes
# the device from provisioning and finally decommissions it. This operation can take a few minutes.
# Supported from CVP 2021.3.0+ and CVaaS.
decomm_status = "DECOMMISSIONING_STATUS_SUCCESS"
decomm_result = ""
while decomm_result != decomm_status:
    decomm_result = clnt.api.device_decommissioning_status_get_one(request_id)['value']['status']
    time.sleep(10)

print(decomm_result)



# Copyright (c) 2021 Arista Networks, Inc.
# Use of this source code is governed by the Apache License 2.0
# that can be found in the COPYING file.

from cvprac.cvp_client import CvpClient
import ssl
ssl._create_default_https_context = ssl._create_unverified_context
import requests.packages.urllib3
requests.packages.urllib3.disable_warnings()
from datetime import datetime

# Note API token auth method is not yet supported with Change Controls
clnt = CvpClient()
clnt.connect(['cvp1'],'username', 'password')

ccid = 'cvprac0904211418'
name = "cvprac CC test"
tlist = ['1021','1020','1019','1018']

### Create Change control with the list of tasks
clnt.api.create_change_control_v3(ccid, name, tlist)

### Approve CC
clnt.api.approve_change_control('cvprac0904211418', timestamp=datetime.utcnow().isoformat() + 'Z')

### Execute CC
clnt.api.execute_change_controls(['cvprac0904211418'])



# Copyright (c) 2021 Arista Networks, Inc.
# Use of this source code is governed by the Apache License 2.0
# that can be found in the COPYING file.
#
# Get list of configlets in parallel

from cvprac.cvp_client import CvpClient
import ssl
from concurrent.futures import ThreadPoolExecutor
ssl._create_default_https_context = ssl._create_unverified_context
import requests.packages.urllib3
requests.packages.urllib3.disable_warnings()

# Create connection to CloudVision
clnt = CvpClient()

clnt.connect(nodes=['cvp1'], username="username",password="password")

import time
from functools import wraps

def get_list_of_configlets(configlets):
    """
    Create a thread pool and download specified urls
    """

    futures_list = []
    results = []

    with ThreadPoolExecutor(max_workers=40) as executor:
        for configlet in configlets:
            futures = executor.submit(clnt.api.get_configlet_by_name, configlet)
            futures_list.append(futures)

        for future in futures_list:
            try:
                result = future.result(timeout=60)
                results.append(result)
            except Exception:
                results.append(None)
    return results

if __name__ == "__main__":
    # Example with pre-defined list
    configlets = ["tp-avd_tp-avd-leaf1","tp-avd_tp-avd-leaf2","tp-avd_tp-avd-leaf3","tp-avd_tp-avd-leaf4"]

    # Example with loading list of configlets from a file
    # with open("configlet_list.txt") as f:
    #     configlets = f.read().splitlines()

    results = get_list_of_configlets(configlets)
    for result in results:
        print(result)



# Copyright (c) 2021 Arista Networks, Inc.
# Use of this source code is governed by the Apache License 2.0
# that can be found in the COPYING file.

# Example on how to re-trigger task creation if a config push task was previously
# cancelled and the device is still config out of sync
import argparse
import ssl
import sys
from packaging.version import parse
from getpass import getpass
from cvprac.cvp_client import CvpClient
import requests.packages.urllib3
requests.packages.urllib3.disable_warnings()


if ((sys.version_info.major == 3) or
        (sys.version_info.major == 2 and sys.version_info.minor == 7 and
         sys.version_info.micro >= 5)):
    ssl._create_default_https_context = ssl._create_unverified_context


def main():
        
    compliance = {"0001": "Config is out of sync",
        "0003": "Config & image out of sync",
        "0004": "Config, Image and Device time are in sync",
        "0005": "Device is not reachable",
        "0008": "Config, Image and Extensions are out of sync",
        "0009": "Config and Extensions are out of sync",
        "0012": "Config, Image, Extension and Device time are out of sync",
        "0013": "Config, Image and Device time are out of sync",
        "0014": "Config, Extensions and Device time are out of sync",
        "0016": "Config and Device time are out of sync"
    }
    # Create connection to CloudVision
    clnt = CvpClient()

    parser = argparse.ArgumentParser(
        description='Script to recreate a task, if a previous config push was cancelled')
    parser.add_argument('-u', '--username', default='username')
    parser.add_argument('-p', '--password', default=None)
    parser.add_argument('-c', '--cvpserver', action='append')
    parser.add_argument('-f', '--filter', action='append', default=None)
    args = parser.parse_args()

    if args.password is None:
        args.password = getpass()

    for cvpserver in args.cvpserver:
        print("Connecting to %s" % cvpserver)
        try:
            clnt.connect(nodes=[cvpserver], username=args.username, password=args.password)
        except Exception as e:
            print("Unable to connect to CVP: %s" % str(e))

        # Get the current CVP version
        cvp_release = clnt.api.get_cvp_info()['version']
        if parse(cvp_release) < parse('2020.3.0'):
            # For older CVP, we manually trigger a compliance check
            try:
                clnt.api.check_compliance('root', 'container')
            except:
                # Bad practice, but the check compliance applied to a container can't actually work
                # since the complianceIndication  key doesn't exist on the container level
                pass
        else:
            # with continuous compliance checks, triggering the check is no longer required
            pass

        device_filters = []
        if args.filter is not None:
            for entry in args.filter:
                device_filters.extend(entry.split(','))

        # Get inventory
        print("Collecting inventory...")
        devices = clnt.api.get_inventory()
        print("%d devices in inventory" % len(devices) )

        for switch in devices:
            if (switch['status'] == 'Registered' and
                    switch['parentContainerId'] != 'undefined_container'):

                if len(device_filters) > 0:
                    # iterate over device filters, and update task for
                    # any devices not in compliance
                    
                    for filter_term in device_filters:
                        print("Checking device: %s" % switch['hostname'])
                        if filter_term in switch['hostname']:
                            # generate configlet list
                            cl = clnt.api.get_configlets_by_device_id(switch['systemMacAddress'])
                            # generate a task if config is out of sync
                            if switch['complianceCode'] in compliance.keys():
                                print(clnt.api.apply_configlets_to_device("", switch, cl))
                            else:
                                print("%s is compliant, nothing to do" % switch['hostname'])
                        else:
                            print("Skipping %s due to filter" % switch['hostname'])
                else:
                    print("Checking device: %s" % switch['hostname'])
                    cl = clnt.api.get_configlets_by_device_id(switch['systemMacAddress'])
                    # generate a task if config is out of sync
                    if switch['complianceCode'] in compliance.keys():
                        print(clnt.api.apply_configlets_to_device("", switch, cl))

            else:
                print("Skipping %s, device is unregistered for provisioning" % switch['hostname'])

    return 0


if __name__ == "__main__":
    main()



from cvprac.cvp_client import CvpClient
import ssl
ssl._create_default_https_context = ssl._create_unverified_context
import requests.packages.urllib3
requests.packages.urllib3.disable_warnings()

# Create connection to CloudVision
clnt = CvpClient()
clnt.connect(['cvp1'],'username', 'password')
oldName = "test"
newName = "test121"

container_id = clnt.api.get_container_by_name(oldName)['key']

data = {"data":[{"info": "Container {} renamed from {}".format(newName, oldName),
                "infoPreview": "Container {} renamed from {}".format(newName, oldName),
                "action": "update",
                "nodeType": "container",
                "nodeId": container_id,
                "toId":"",
                "fromId":"",
                "nodeName": newName,
                "fromName": "",
                "toName": "",
                "toIdType": "container",
                "oldNodeName": oldName
                }
            ]
        }

clnt.api._add_temp_action(data)
clnt.api._save_topology_v2([])



# Copyright (c) 2021 Arista Networks, Inc.
# Use of this source code is governed by the Apache License 2.0
# that can be found in the COPYING file.

from cvprac.cvp_client import CvpClient
import ssl
ssl._create_default_https_context = ssl._create_unverified_context
import requests.packages.urllib3
requests.packages.urllib3.disable_warnings()

# Create connection to CloudVision
clnt = CvpClient()
clnt.connect(nodes=['cvp1'], username="username",password="password")

configletNames = ['tp-avd_tp-avd-leaf1','vlan144','api_models']

device_name = "tp-avd-leaf1"
device = clnt.api.get_device_by_name(device_name)

configlets = []

for name in configletNames:
    configlets.append(clnt.api.get_configlet_by_name(name))

# Apply configlets in the order specified in the list
clnt.api.apply_configlets_to_device("", device, configlets, reorder_configlets=True)



# Copyright (c) 2021 Arista Networks, Inc.
# Use of this source code is governed by the Apache License 2.0
# that can be found in the COPYING file.

from cvprac.cvp_client import CvpClient
import ssl
ssl._create_default_https_context = ssl._create_unverified_context
import requests.packages.urllib3
requests.packages.urllib3.disable_warnings()

# Create connection to CloudVision
clnt = CvpClient()
clnt.connect(nodes=['cvp1'], username="username",password="password")

inventory = clnt.api.get_inventory()

devices = []
for netelement in inventory:
   devices.append(netelement['systemMacAddress'])

# Remove devices from provisioning
# This is a legacy API call that removes the devices from Network Provisioning 
# in CVP versions older than 2021.3.0, however it does not remove them from 
# the Device Inventory as that requires the streaming agent (TerminAttr) to be shutdown,
# which this API does not support.
# To fully decommission a device the device_decommissioning() API can be used, which is
# supported from 2021.3.0+.
# Note that using the delete_devices() function post CVP 2021.3.0 the device will be 
# automatically added back to the Undefined container.
clnt.api.delete_devices(devices)



# Copyright (c) 2021 Arista Networks, Inc.
# Use of this source code is governed by the Apache License 2.0
# that can be found in the COPYING file.
#
# Example script to generate the TerminAttr token via REST API from CVaaS and CV on-prem
# and save them to a file

from cvprac.cvp_client import CvpClient
from pprint import pprint as pp
import ssl
ssl._create_default_https_context = ssl._create_unverified_context
import requests.packages.urllib3
requests.packages.urllib3.disable_warnings()

# Reading the service account token from a file
with open("cvaas.tok") as f:
    token = f.read().strip('\n')

clnt = CvpClient()
clnt.connect(nodes=['www.arista.io'], username='',password='',is_cvaas=True, api_token=token)

terminattr_token = clnt.api.create_enroll_token('720h')
with open('cv-onboarding-token', 'w') as f:
    f.write(terminattr_token[0]['enrollmentToken']['token'])

primary = CvpClient()
primary.connect(nodes=['cvp1'], username='username',password='password')

terminattr_token = primary.api.create_enroll_token('720h')

with open('token', 'w') as f:
    f.write(terminattr_token['data'])



# Copyright (c) 2021 Arista Networks, Inc.
# Use of this source code is governed by the Apache License 2.0
# that can be found in the COPYING file.

from cvprac.cvp_client import CvpClient
from cvprac.cvp_client_errors import CvpApiError
import ssl
ssl._create_default_https_context = ssl._create_unverified_context
import requests.packages.urllib3
requests.packages.urllib3.disable_warnings()
from cvprac.cvp_client import CvpClient
import csv

# Create connection to CloudVision using Service Account token
with open("cvaas.tok") as f:
    token = f.read().strip('\n')

clnt = CvpClient()
clnt.connect(nodes=['www.arista.io'], username='', password='', is_cvaas=True, api_token=token)


with open("aaa_users.csv") as csvfile:
    for i in csv.DictReader(csvfile):
        data = dict(i)
        try:
            clnt.api.add_user(data['username'], "", data['role'], data['status'], data['first_name'], data['last_name'], data['email'], data['user_type'])
        except CvpApiError as e:
            print(e)
        print ("Adding user {} to CVaaS".format(data['username']))



# Copyright (c) 2020 Arista Networks, Inc.
# Use of this source code is governed by the Apache License 2.0
# that can be found in the COPYING file.

from cvprac.cvp_client import CvpClient
import ssl
ssl._create_default_https_context = ssl._create_unverified_context
import requests.packages.urllib3
requests.packages.urllib3.disable_warnings()

# Create connection to CloudVision
clnt = CvpClient()
clnt.connect(['cvp1'],'username', 'password')

container_id = clnt.api.get_container_by_name("TP_LEAFS")['key']
builder_name = 'SYS_TelemetryBuilderV3'
configletBuilderID = clnt.api.get_configlet_by_name(builder_name)['key']

payload = {"previewValues":[{
                            "fieldId":"vrf",
                            "value":"red"}],
            "configletBuilderId":configletBuilderID,
            "netElementIds":[],
            "pageType":"container",
            "containerId":container_id,
            "containerToId":"",
            "mode":"assign"}

preview = clnt.post('/configlet/configletBuilderPreview.do', data=payload)

generated_names_list = []
generated_keys_list = []

for i in preview['data']:
  generated_names_list.append(i['configlet']['name'])
  generated_keys_list.append(i['configlet']['key'])

clnt.get("/configlet/searchConfiglets.do?objectId={}&objectType=container&type=ignoreDraft&queryparam={}&startIndex=0&endIndex=22&sortByColumn=&sortOrder=".format(container_id, builder_name.lower()))

tempData = {"data":[{
    "info":"Configlet Assign: to container TP_LEAFS",
    "infoPreview":"<b>Configlet Assign:</b> to container  TP_LEAFS",
    "action":"associate",
    "nodeType":"configlet",
    "nodeId":"",
    "toId":container_id,
    "fromId":"","nodeName":"","fromName":"",
    "toName":"TP_LEAFS",
    "toIdType":"container",
    "configletList":generated_keys_list,
    "configletNamesList":generated_names_list,
    "ignoreConfigletList":[],
    "ignoreConfigletNamesList":[],
    "configletBuilderList":[configletBuilderID],
    "configletBuilderNamesList":[builder_name],
    "ignoreConfigletBuilderList":[],
    "ignoreConfigletBuilderNamesList":[]
    }
    ]
    }

clnt.api._add_temp_action(tempData)
clnt.api._save_topology_v2([])



# Copyright (c) 2021 Arista Networks, Inc.
# Use of this source code is governed by the Apache License 2.0
# that can be found in the COPYING file.

from cvprac.cvp_client import CvpClient
import ssl
ssl._create_default_https_context = ssl._create_unverified_context
import requests.packages.urllib3
requests.packages.urllib3.disable_warnings()

# Create connection to CloudVision
clnt = CvpClient()
clnt.connect(nodes=['cvp1'], username="username",password="password")

# Modify existing configlet

configletName = "cvprac_example"

configlet = """!
interface Ethernet10
   description DUB_R04
   ip address 10.144.144.2/24
!
"""

configletID = clnt.api.get_configlet_by_name(configletName)['key']

clnt.api.update_configlet( configlet, configletID, configletName)



# Copyright (c) 2021 Arista Networks, Inc.
# Use of this source code is governed by the Apache License 2.0
# that can be found in the COPYING file.
#
# Get configlets and save them to individual files
from cvprac.cvp_client import CvpClient
import ssl
ssl._create_default_https_context = ssl._create_unverified_context
import requests.packages.urllib3
requests.packages.urllib3.disable_warnings()

# Create connection to CloudVision
clnt = CvpClient()
clnt.connect(['cvp1'],'username', 'password')

configlets = clnt.api.get_configlets(start=0,end=0)

for configlet in configlets['data']:
    with open(configlet['name'],'w') as f:
        f.write(configlet['config'])



#!/usr/bin/env python3
#
# python3 mlag_issu <upgrade inventory file> <MLAG peer to upgrade: 'peer1' or 'peer2'>"
#
# # Example of upgrade inventory file (YAML)
# cvp_hosts:
#     - 192.168.0.191
#     - 192.168.0.192
#     - 192.168.0.193
# cvp_username: cvpadmin
# target_eos_version: 4.25.4M
# target_terminattr_version: 1.13.6
# mlag_couples:
#     - peer1: leaf101-1
#       peer2: leaf101-2
#     - peer1: leaf102-1
#       peer2: leaf102-2
#
# Note: upgrades are performed in parallel

import sys
import time
import string
import random
from getpass import getpass
import requests
from requests.packages.urllib3.exceptions import InsecureRequestWarning
from datetime import datetime
from cvprac.cvp_client import CvpClient
from cvprac.cvp_client_errors import CvpLoginError, CvpApiError
from pprint import pprint
from operator import itemgetter
import yaml

class CvpDeviceUpgrader(object):
    def __init__(self, hosts, username, password):
        requests.packages.urllib3.disable_warnings(InsecureRequestWarning)
        self.cvp_hosts = hosts
        self.cvp_user = username
        self.cvp_password = password
        self.session = self._open_cvp_session()

    def _open_cvp_session(self):
        try:
            client = CvpClient()
            client.connect(
                nodes=self.cvp_hosts,
                username=self.cvp_user,
                password=self.cvp_password,
                request_timeout=300,
                connect_timeout=30
            )
            return(client)
        except CvpLoginError as e:
            print(f"Cannot connect to CVP API: {e}")
            exit()

    def create_mlag_issu_change_control(self, taskIDs, deviceIDs):
        cc_id = f"CC_{datetime.now().strftime('%Y%m%d_%H%M%S')}"
        pre_upgrade_stage = {'stage': [{
            'id': f"preU_{cc_id}",
            'name': 'pre_upgrade',
            'stage_row':[{'stage': [{
                'id': ''.join(random.choice(string.ascii_letters + string.digits) for _ in range(9)),
                'action': {
                    'name': 'mlaghealthcheck',
                    'timeout': 0,
                    'args': {
                        'DeviceID': device_id
                    }
                }
            } for device_id in deviceIDs]}]
        }]}
        upgrade_stage = {'stage': [{
            'id': f"U_{cc_id}",
            'name': 'upgrade',
            'stage_row': [{'stage': [{
                'id': task_id,
                'action': {
                    'name': 'task',
                    'args': {
                        'TaskID': task_id
                    }
                }
            } for task_id in taskIDs]}]
        }]}
        post_upgrade_stage = {'stage': [{
            'id': f"postU_{cc_id}",
            'name': 'post_upgrade',
            'stage_row': [{'stage': [{
                'id': ''.join(random.choice(string.ascii_letters + string.digits) for _ in range(9)),
                'action': {
                    'name': 'mlaghealthcheck',
                    'timeout': 0,
                    'args': {
                        'DeviceID': device_id
                    }
                }
            } for device_id in deviceIDs]}]
        }]}
        cc_data = {'config': {
            'id': cc_id,
            'name': f"Change Control {cc_id}",
            'root_stage': {
                'id': 'root',
                'name': f"Change Control {cc_id} root",
                'stage_row': [pre_upgrade_stage, upgrade_stage, post_upgrade_stage],
            }
        }}
        try:
            res = self.session.post('/api/v3/services/ccapi.ChangeControl/Update',
                data=cc_data,
                timeout=self.session.api.request_timeout
            )
        except Exception as e:
            print(str(e))
            return(None)
        print(f"Change control {res[0]['id']} created at {res[0]['update_timestamp']}")
        return(res[0]['id'])

    def get_mlag_issu_change_control_logs(self, ccID, startTime):
        end_time = int(time.time() * 1000)
        cc_logs_data = {'category': 'ChangeControl',
            'objectKey': ccID,
            'dataSize': 15000,
            'startTime': startTime,
            'endTime': end_time
        }
        logs = self.session.post('/cvpservice/audit/getLogs.do',
            data=cc_logs_data,
            timeout=self.session.api.request_timeout
        )
        for log in sorted(logs['data'], key=itemgetter('dateTimeInLongFormat')):
            if log['subObjectName'] and 'Command(s)' not in log['activity']:
                log_date = datetime.fromtimestamp(log['dateTimeInLongFormat']/1000)
                print(f"{log_date} {log['subObjectName']}: {log['activity']}")
        return(end_time + 1)

    def run_mlag_issu_change_control(self, ccID):
        print(f"Automatic approval of change control {ccID}")
        self.session.api.approve_change_control(ccID, datetime.utcnow().isoformat() + 'Z')
        time.sleep(2)
        print(f"Starting the execution of change control {ccID}")
        start_time = int(time.time() * 1000)
        self.session.api.execute_change_controls([ccID])
        time.sleep(2)
        cc_status = self.session.api.get_change_control_status(ccID)[0]['status']
        start_time = self.get_mlag_issu_change_control_logs(ccID, start_time)
        while cc_status['state'] == 'Running':
            time.sleep(30)
            cc_status = self.session.api.get_change_control_status(ccID)[0]['status']
            start_time = self.get_mlag_issu_change_control_logs(ccID, start_time)
        print(f"Change control {ccID} final status: {cc_status['state']}")
        if cc_status['error']:
            print(f"Change control {ccID} had the following errors: {cc_status['error']}")
        else:
            print(f"Change control {ccID} completed without errors")

def main():
    if len(sys.argv) != 3:
        print(f"Usage: python3 {sys.argv[0]} <input file path> <MLAG peer to upgrade: peer1/peer2>")
        exit()
    try:
        with open(sys.argv[1], 'r') as yf:
            params = yaml.safe_load(yf)
    except Exception as e:
        print(e)
        exit()
    cvp_password = getpass(prompt=f"CVP password for user {params['cvp_username']}: ")
    cvpdu = CvpDeviceUpgrader(
        hosts=params['cvp_hosts'],
        username=params['cvp_username'],
        password=cvp_password
    )
    image_bundle = None
    for bundle in cvpdu.session.api.get_image_bundles()['data']:
        eos_match = False
        terminattr_match = False
        for img in bundle['imageIds']:
            if params['target_eos_version'] in img:
                eos_match = True
            elif params['target_terminattr_version'] in img:
                terminattr_match = True
        if eos_match and terminattr_match:
            image_bundle = bundle
            break
    if image_bundle is None:
        print(f"Cannot find an image bundle with EOS {params['target_eos_version']} and TerminAttr {params['target_terminattr_version']}")
        exit()
    hostnames = [couple[sys.argv[2]] for couple in params['mlag_couples']]
    devices_to_upgrade = list()
    inventory = cvpdu.session.api.get_inventory()
    for hostname in hostnames:
        provisioned = False
        for dev in inventory:
            if dev['hostname'] == hostname:
                provisioned = True
                devices_to_upgrade.append(dev)
                break
        if not provisioned:
            print(f"Device with hostname {hostname} is not provisioned in CVP")
    if not devices_to_upgrade:
        print('none of the mentioned devices is provisioned in CVP')
        exit()
    print(f"Devices to upgrade: {', '.join([dev['hostname'] for dev in devices_to_upgrade])}")
    task_ids = list()
    for device in devices_to_upgrade:
        response = cvpdu.session.api.apply_image_to_device(image_bundle, device)['data']
        if response['status'] == 'success':
            task_ids.extend(response['taskIds'])
    device_ids = [device['serialNumber'] for device in devices_to_upgrade]
    cc_id = cvpdu.create_mlag_issu_change_control(task_ids, device_ids)
    if cc_id is None:
        print('Failed to create the MLAG ISSU change control')
        exit()
    time.sleep(2)
    cvpdu.run_mlag_issu_change_control(cc_id)

if __name__ == '__main__':
    main()



# Copyright (c) 2021 Arista Networks, Inc.
# Use of this source code is governed by the Apache License 2.0
# that can be found in the COPYING file.
#
# NOTE: The following example is using the new Change Control Resource APIs supported in 2021.2.0 or newer and in CVaaS.
# For CVaaS service-account token based auth has to be used.

from cvprac.cvp_client import CvpClient
import ssl
import uuid
from datetime import datetime
ssl._create_default_https_context = ssl._create_unverified_context
import requests.packages.urllib3
requests.packages.urllib3.disable_warnings()

# Create connection to CloudVision
clnt = CvpClient()
clnt.connect(['cvp1'],'username', 'password')


cc_id = str(uuid.uuid4())
name = f"Change_{datetime.now().strftime('%Y%m%d_%H%M%S')}"

# Create custom stage hierarchy
# The below example would result in the following hierarchy:
# root (series)
# |- stages 1-2 (series)
# |  |- stage 1ab (parallel)
# |  |    |- stage 1a
# |  |    |- stage 1b
# |  |- stage 2
# |- stage 3
data = {'key': {
            'id': cc_id
            },
        'change': {
            'name': cc_id,
            'notes': 'cvprac CC',
            'rootStageId': 'root',
            'stages': {'values': {'root': {'name': 'root',
                                           'rows': {'values': [{'values': ['1-2']},
                                                               {'values': ['3']}]
                                            }
                                        },
                                  '1-2': {'name': 'stages 1-2',
                                          'rows': {'values': [{'values': ['1ab']},
                                                              {'values': ['2']}]}},
                                  '1ab': {'name': 'stage 1ab',
                                          'rows': {'values': [{'values': ['1a','1b']}]
                                            }
                                        },
                                  '1a': {'action': {'args': {'values': {'TaskID': '1242'}},
                                                    'name': 'task',
                                                    'timeout': 3000},
                                         'name': 'stage 1a'},
                                  '1b': {'action': {'args': {'values': {'TaskID': '1243'}},
                                                    'name': 'task',
                                                    'timeout': 3000},
                                         'name': 'stage 1b'},
                                  '2': {'action': {'args': {'values': {'TaskID': '1240'}},
                                                   'name': 'task',
                                                   'timeout': 3000},
                                        'name': 'stage 2'},
                                  '3': {'action': {'args': {'values': {'TaskID': '1241'}},
                                                   'name': 'task',
                                                   'timeout': 3000},
                                        'name': 'stage 3'},
                }
            }
        }
 }
# Create change control from custom stage hierarchy data
clnt.api.change_control_create_with_custom_stages(data)

# Approve the change control
approval_note = "Approve CC via cvprac" # notes are optional
clnt.api.change_control_approve(cc_id, notes=approval_note)

# Start the change control
start_note = "Starting CC via cvprac" # notes are optional
clnt.api.change_control_start(cc_id, notes=start_note)
